"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Naming_1 = require("./Naming");
var Renderer_1 = require("./Renderer");
var Support_1 = require("./Support");
var ConvenienceRenderer = /** @class */ (function (_super) {
    __extends(ConvenienceRenderer, _super);
    function ConvenienceRenderer() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.nameForTopLevel = function (type, name) {
            var maybeNamedType = _this.namedTypeToNameForTopLevel(type);
            var styledName;
            if (maybeNamedType) {
                styledName = _this.namedTypeNamer.nameStyle(name);
            }
            else {
                styledName = _this.topLevelNameStyle(name);
            }
            var named = _this.globalNamespace.add(new Naming_1.FixedName(styledName));
            var dependencyNames = _this.topLevelDependencyNames(named);
            for (var _i = 0, dependencyNames_1 = dependencyNames; _i < dependencyNames_1.length; _i++) {
                var dn = dependencyNames_1[_i];
                _this.globalNamespace.add(dn);
            }
            if (maybeNamedType) {
                _this._classAndUnionNames = _this._classAndUnionNames.set(maybeNamedType, named);
            }
            return named;
        };
        _this.addClassOrUnionNamed = function (type) {
            var existing = _this._classAndUnionNames.get(type);
            if (existing !== undefined)
                return existing;
            var name = type.names.combined;
            var named = _this.globalNamespace.add(new Naming_1.SimpleName(name, _this.namedTypeNamer));
            _this._classAndUnionNames = _this._classAndUnionNames.set(type, named);
            return named;
        };
        _this.addPropertyNameds = function (c, classNamed) {
            var _a = _this.forbiddenForProperties(c, classNamed), forbiddenNames = _a.names, forbiddenNamespace = _a.namespaces;
            var ns = new Naming_1.Namespace(c.names.combined, _this.globalNamespace, immutable_1.Set(forbiddenNamespace), immutable_1.Set(forbiddenNames));
            var names = c.properties
                .map(function (t, name) {
                return ns.add(new Naming_1.SimpleName(name, _this.propertyNamer));
            })
                .toMap();
            _this._propertyNames = _this._propertyNames.set(c, names);
        };
        _this.childrenOfType = function (t) {
            var names = _this.names;
            if (t instanceof Type_1.ClassType) {
                var propertyNameds_1 = Support_1.defined(_this._propertyNames.get(t));
                return t.properties
                    .sortBy(function (_, n) {
                    return Support_1.defined(names.get(Support_1.defined(propertyNameds_1.get(n))));
                })
                    .toOrderedSet();
            }
            return t.children.toOrderedSet();
        };
        _this.unionFieldName = function (t) {
            var typeNameForUnionMember = function (t) {
                if (t instanceof Type_1.PrimitiveType) {
                    switch (t.kind) {
                        case "any":
                            return "anything";
                        case "null":
                            return "null";
                        case "bool":
                            return "bool";
                        case "integer":
                            return "long";
                        case "double":
                            return "double";
                        case "string":
                            return "string";
                    }
                }
                else if (t instanceof Type_1.ArrayType) {
                    return typeNameForUnionMember(t.items) + "_array";
                }
                else if (t instanceof Type_1.ClassType) {
                    return Support_1.defined(_this.names.get(_this.nameForNamedType(t)));
                }
                else if (t instanceof Type_1.MapType) {
                    return typeNameForUnionMember(t.values), "_map";
                }
                else if (t instanceof Type_1.UnionType) {
                    return "union";
                }
                throw "Unknown type";
            };
            return _this.propertyNamer.nameStyle(typeNameForUnionMember(t));
        };
        _this.nameForNamedType = function (t) {
            if (!_this._classAndUnionNames.has(t)) {
                throw "Named type does not exist.";
            }
            return Support_1.defined(_this._classAndUnionNames.get(t));
        };
        _this.forEachTopLevel = function (blankLocations, f) {
            _this.forEachWithBlankLines(_this.topLevels, blankLocations, function (t, name) {
                return f(t, Support_1.defined(_this._topLevelNames.get(name)));
            });
        };
        _this.forEachProperty = function (c, blankLocations, f) {
            var propertyNames = Support_1.defined(_this._propertyNames.get(c));
            var sortedPropertyNames = propertyNames
                .sortBy(function (n) { return _this.names.get(n); })
                .toOrderedMap();
            _this.forEachWithBlankLines(sortedPropertyNames, blankLocations, function (name, jsonName) {
                var t = Support_1.defined(c.properties.get(jsonName));
                f(name, jsonName, t);
            });
        };
        _this.callForClass = function (c, f) {
            f(c, Support_1.defined(_this._classAndUnionNames.get(c)));
        };
        _this.forEachClass = function (blankLocations, f) {
            _this.forEachWithBlankLines(_this._namedClasses, blankLocations, function (c) {
                _this.callForClass(c, f);
            });
        };
        _this.callForUnion = function (u, f) {
            f(u, Support_1.defined(_this._classAndUnionNames.get(u)));
        };
        _this.forEachUnion = function (blankLocations, f) {
            _this.forEachWithBlankLines(_this._namedUnions, blankLocations, function (u) { return _this.callForUnion(u, f); });
        };
        _this.forEachUniqueUnion = function (blankLocations, unionMembers, f) {
            var uniqueUnions = _this._namedUnions.map(unionMembers);
            _this.forEachWithBlankLines(uniqueUnions, blankLocations, f);
        };
        _this.forEachNamedType = function (blankLocations, leavesFirst, classFunc, unionFunc) {
            var collection = _this._namedTypes;
            if (leavesFirst)
                collection = collection.reverse();
            _this.forEachWithBlankLines(collection, blankLocations, function (t) {
                if (t instanceof Type_1.ClassType) {
                    _this.callForClass(t, classFunc);
                }
                else if (t instanceof Type_1.UnionType) {
                    _this.callForUnion(t, unionFunc);
                }
                else {
                    throw "Named type that's neither a class nor union";
                }
            });
        };
        return _this;
    }
    Object.defineProperty(ConvenienceRenderer.prototype, "forbiddenNamesForGlobalNamespace", {
        get: function () {
            return [];
        },
        enumerable: true,
        configurable: true
    });
    ConvenienceRenderer.prototype.forbiddenForProperties = function (c, classNamed) {
        return { names: [], namespaces: [] };
    };
    ConvenienceRenderer.prototype.topLevelDependencyNames = function (topLevelName) {
        return [];
    };
    ConvenienceRenderer.prototype.unionNeedsName = function (u) {
        return !Type_1.nullableFromUnion(u);
    };
    ConvenienceRenderer.prototype.setUpNaming = function () {
        var _this = this;
        this.globalNamespace = Naming_1.keywordNamespace("global", this.forbiddenNamesForGlobalNamespace);
        var _a = Type_1.allClassesAndUnions(this.topLevels), classes = _a.classes, unions = _a.unions;
        var namedUnions = unions.filter(function (u) { return _this.unionNeedsName(u); }).toOrderedSet();
        this._classAndUnionNames = immutable_1.Map();
        this._propertyNames = immutable_1.Map();
        this._topLevelNames = this.topLevels.map(this.nameForTopLevel).toMap();
        classes.forEach(function (c) {
            var named = _this.addClassOrUnionNamed(c);
            _this.addPropertyNameds(c, named);
        });
        namedUnions.forEach(function (u) { return _this.addClassOrUnionNamed(u); });
        return [this.globalNamespace];
    };
    Object.defineProperty(ConvenienceRenderer.prototype, "namedUnions", {
        get: function () {
            return this._namedUnions;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ConvenienceRenderer.prototype, "haveUnions", {
        get: function () {
            return !this._namedUnions.isEmpty();
        },
        enumerable: true,
        configurable: true
    });
    ConvenienceRenderer.prototype.emitSource = function () {
        var _this = this;
        var types = Type_1.allNamedTypes(this.topLevels, this.childrenOfType);
        this._namedTypes = types
            .filter(function (t) { return !(t instanceof Type_1.UnionType) || _this.unionNeedsName(t); })
            .toOrderedSet();
        var _a = Type_1.splitClassesAndUnions(this._namedTypes), classes = _a.classes, unions = _a.unions;
        this._namedClasses = classes;
        this._namedUnions = unions;
        this.emitSourceStructure();
    };
    return ConvenienceRenderer;
}(Renderer_1.Renderer));
exports.ConvenienceRenderer = ConvenienceRenderer;
