"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var unicode = require("unicode-properties");
// FIXME: This is a copy of code in src/Data/String/Util.js
function stringConcatMap(mapper) {
    var charStringMap = [];
    var charNoEscapeMap = [];
    for (var i = 0; i < 128; i++) {
        var noEscape = 0;
        var input = String.fromCharCode(i);
        var result = mapper(input);
        if (result === input) {
            noEscape = 1;
        }
        charStringMap.push(result);
        charNoEscapeMap.push(noEscape);
    }
    return function stringConcatMap_inner(s) {
        var cs = null;
        var start = 0;
        var i = 0;
        while (i < s.length) {
            var cc = s.charCodeAt(i);
            if (!charNoEscapeMap[cc]) {
                if (cs === null)
                    cs = [];
                cs.push(s.substring(start, i));
                var str = charStringMap[cc];
                if (str === undefined) {
                    cs.push(mapper(s.charAt(i)));
                }
                else {
                    cs.push(str);
                }
                start = i + 1;
            }
            i++;
        }
        if (cs === null)
            return s;
        cs.push(s.substring(start, i));
        return cs.join("");
    };
}
exports.stringConcatMap = stringConcatMap;
function legalizeCharacters(isLegal) {
    return stringConcatMap(function (c) { return (isLegal(c) ? c : "_"); });
}
exports.legalizeCharacters = legalizeCharacters;
function intToHex(i, width) {
    var str = i.toString(16);
    if (str.length >= width)
        return str;
    return "0".repeat(width - str.length) + str;
}
function standardUnicodeHexEscape(c) {
    var i = c.charCodeAt(0);
    if (i <= 0xffff) {
        return "\\u" + intToHex(i, 4);
    }
    else {
        return "\\U" + intToHex(i, 8);
    }
}
exports.standardUnicodeHexEscape = standardUnicodeHexEscape;
function genericStringEscape(escaper) {
    function mapper(c) {
        switch (c) {
            case "\\":
                return "\\\\";
            case '"':
                return '\\"';
            case "\n":
                return "\\n";
            case "\t":
                return "\\t";
            default:
                if (isPrintable(c)) {
                    return c;
                }
                return escaper(c);
        }
    }
    return stringConcatMap(mapper);
}
exports.stringEscape = genericStringEscape(standardUnicodeHexEscape);
function isPrintable(c) {
    var category = unicode.getCategory(c.charCodeAt(0));
    return ([
        "Mc",
        "No",
        "Sk",
        "Me",
        "Nd",
        "Po",
        "Lt",
        "Pc",
        "Sm",
        "Zs",
        "Lu",
        "Pd",
        "So",
        "Pe",
        "Pf",
        "Ps",
        "Sc",
        "Ll",
        "Lm",
        "Pi",
        "Nl",
        "Mn",
        "Lo"
    ].indexOf(category) >= 0);
}
function isLetter(c) {
    var category = unicode.getCategory(c.charCodeAt(0));
    // FIXME: Include Letter, modifier (Lm)?
    return ["Lu", "Ll", "Lt", "Lo"].indexOf(category) >= 0;
}
exports.isLetter = isLetter;
function isDigit(c) {
    var category = unicode.getCategory(c.charCodeAt(0));
    return ["Nd"].indexOf(category) >= 0;
}
exports.isDigit = isDigit;
function isLetterOrUnderscore(c) {
    return isLetter(c) || c == "_";
}
exports.isLetterOrUnderscore = isLetterOrUnderscore;
function isLetterOrUnderscoreOrDigit(c) {
    return isLetter(c) || isDigit(c) || c == "_";
}
exports.isLetterOrUnderscoreOrDigit = isLetterOrUnderscoreOrDigit;
function modifyFirstChar(f, s) {
    if (s === "")
        return s;
    return f(s[0]) + s.slice(1);
}
function capitalize(str) {
    return modifyFirstChar(function (c) { return c.toUpperCase(); }, str);
}
exports.capitalize = capitalize;
function decapitalize(str) {
    return modifyFirstChar(function (c) { return c.toLowerCase(); }, str);
}
exports.decapitalize = decapitalize;
var wordSeparatorRegex = RegExp("[-_. ]");
function camelCase(str) {
    var words = str.split(wordSeparatorRegex).map(capitalize);
    return words.join("");
}
exports.camelCase = camelCase;
function startWithLetter(isLetter, upper, str) {
    var modify = upper ? capitalize : decapitalize;
    if (str === "")
        return modify("empty");
    if (isLetter(str[0]))
        return modify(str);
    return modify("the" + str);
}
exports.startWithLetter = startWithLetter;
function intercalate(separator, items) {
    var acc = [];
    items.forEach(function (x) {
        if (acc.length > 0)
            acc.push(separator);
        acc.push(x);
    });
    return immutable_1.List(acc);
}
exports.intercalate = intercalate;
function defined(x) {
    if (x !== undefined)
        return x;
    throw "Defined value expected, but got undefined";
}
exports.defined = defined;
function nonNull(x) {
    if (x !== null)
        return x;
    throw "Non-null value expected, but got null";
}
exports.nonNull = nonNull;
