import { OrderedSet, Map, Set, Collection } from "immutable";
import { TypeKind, PrimitiveTypeKind, NamedTypeKind } from "./Reykjavik"
export declare type TypeNames = {
    names: Set<string>;
    combined: string;
};
export declare type TopLevels = Map<string, Type>;
export declare abstract class Type {
    readonly kind: TypeKind;
    constructor(kind: TypeKind);
    isNamedType(): this is NamedType;
    readonly abstract children: OrderedSet<Type>;
    readonly directlyReachableNamedTypes: OrderedSet<NamedType>;
    abstract equals(other: any): boolean;
    abstract hashCode(): number;
}
export declare class PrimitiveType extends Type {
    readonly kind: PrimitiveTypeKind;
    constructor(kind: PrimitiveTypeKind);
    readonly children: OrderedSet<Type>;
    equals(other: any): boolean;
    hashCode(): number;
}
export declare class ArrayType extends Type {
    readonly items: Type;
    readonly kind: "array";
    constructor(items: Type);
    readonly children: OrderedSet<Type>;
    equals(other: any): boolean;
    hashCode(): number;
}
export declare class MapType extends Type {
    readonly values: Type;
    readonly kind: "map";
    constructor(values: Type);
    readonly children: OrderedSet<Type>;
    equals(other: any): boolean;
    hashCode(): number;
}
export declare abstract class NamedType extends Type {
    readonly names: TypeNames;
    constructor(kind: NamedTypeKind, names: TypeNames);
    isNamedType(): this is NamedType;
}
export declare class ClassType extends NamedType {
    kind: "class";
    private _properties;
    constructor(names: TypeNames);
    setProperties(properties: Map<string, Type>): void;
    readonly properties: Map<string, Type>;
    readonly children: OrderedSet<Type>;
    equals(other: any): boolean;
    hashCode(): number;
}
export declare class UnionType extends NamedType {
    readonly members: OrderedSet<Type>;
    kind: "union";
    constructor(names: TypeNames, members: OrderedSet<Type>);
    findMember: (kind: TypeKind) => Type | undefined;
    readonly children: OrderedSet<Type>;
    equals(other: any): boolean;
    hashCode(): number;
}
export declare function removeNullFromUnion(t: UnionType): [boolean, OrderedSet<Type>];
export declare function nullableFromUnion(t: UnionType): Type | null;
export declare function allNamedTypes(graph: TopLevels, childrenOfType?: (t: Type) => Collection<any, Type>): OrderedSet<NamedType>;
export declare type ClassesAndUnions = {
    classes: OrderedSet<ClassType>;
    unions: OrderedSet<UnionType>;
};
export declare function splitClassesAndUnions(types: Collection<any, NamedType>): ClassesAndUnions;
export declare function allClassesAndUnions(graph: TopLevels, childrenOfType?: (t: Type) => Collection<any, Type>): ClassesAndUnions;
export declare function matchType<U>(t: Type, anyType: (anyType: PrimitiveType) => U, nullType: (nullType: PrimitiveType) => U, boolType: (boolType: PrimitiveType) => U, integerType: (integerType: PrimitiveType) => U, doubleType: (doubleType: PrimitiveType) => U, stringType: (stringType: PrimitiveType) => U, arrayType: (arrayType: ArrayType) => U, classType: (classType: ClassType) => U, mapType: (mapType: MapType) => U, unionType: (unionType: UnionType) => U): U;
