"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var stringHash = require("string-hash");
var Support_1 = require("./Support");
var Type = /** @class */ (function () {
    function Type(kind) {
        this.kind = kind;
    }
    Type.prototype.isNamedType = function () {
        return false;
    };
    Object.defineProperty(Type.prototype, "directlyReachableNamedTypes", {
        get: function () {
            if (this.isNamedType())
                return immutable_1.OrderedSet([this]);
            return orderedSetUnion(this.children.map(function (t) { return t.directlyReachableNamedTypes; }));
        },
        enumerable: true,
        configurable: true
    });
    return Type;
}());
exports.Type = Type;
var PrimitiveType = /** @class */ (function (_super) {
    __extends(PrimitiveType, _super);
    function PrimitiveType(kind) {
        return _super.call(this, kind) || this;
    }
    Object.defineProperty(PrimitiveType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    PrimitiveType.prototype.equals = function (other) {
        if (!(other instanceof PrimitiveType))
            return false;
        return this.kind === other.kind;
    };
    PrimitiveType.prototype.hashCode = function () {
        return stringHash(this.kind) | 0;
    };
    return PrimitiveType;
}(Type));
exports.PrimitiveType = PrimitiveType;
function isNull(t) {
    return t.kind === "null";
}
var ArrayType = /** @class */ (function (_super) {
    __extends(ArrayType, _super);
    function ArrayType(items) {
        var _this = _super.call(this, "array") || this;
        _this.items = items;
        return _this;
    }
    Object.defineProperty(ArrayType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet([this.items]);
        },
        enumerable: true,
        configurable: true
    });
    ArrayType.prototype.equals = function (other) {
        if (!(other instanceof ArrayType))
            return false;
        return this.items.equals(other.items);
    };
    ArrayType.prototype.hashCode = function () {
        return (stringHash(this.kind) + this.items.hashCode()) | 0;
    };
    return ArrayType;
}(Type));
exports.ArrayType = ArrayType;
var MapType = /** @class */ (function (_super) {
    __extends(MapType, _super);
    function MapType(values) {
        var _this = _super.call(this, "map") || this;
        _this.values = values;
        return _this;
    }
    Object.defineProperty(MapType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet([this.values]);
        },
        enumerable: true,
        configurable: true
    });
    MapType.prototype.equals = function (other) {
        if (!(other instanceof MapType))
            return false;
        return this.values.equals(other.values);
    };
    MapType.prototype.hashCode = function () {
        return (stringHash(this.kind) + this.values.hashCode()) | 0;
    };
    return MapType;
}(Type));
exports.MapType = MapType;
var NamedType = /** @class */ (function (_super) {
    __extends(NamedType, _super);
    function NamedType(kind, names) {
        var _this = _super.call(this, kind) || this;
        _this.names = names;
        return _this;
    }
    NamedType.prototype.isNamedType = function () {
        return true;
    };
    return NamedType;
}(Type));
exports.NamedType = NamedType;
var ClassType = /** @class */ (function (_super) {
    __extends(ClassType, _super);
    function ClassType(names) {
        return _super.call(this, "class", names) || this;
    }
    ClassType.prototype.setProperties = function (properties) {
        if (this._properties !== undefined) {
            throw "Can only set class properties once";
        }
        this._properties = properties;
    };
    Object.defineProperty(ClassType.prototype, "properties", {
        get: function () {
            if (this._properties === undefined) {
                throw "Class properties accessed before they were set";
            }
            return this._properties;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "children", {
        get: function () {
            return this.properties.toOrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    ClassType.prototype.equals = function (other) {
        if (!(other instanceof ClassType))
            return false;
        return (this.names.names.equals(other.names.names) && this.properties.equals(other.properties));
    };
    ClassType.prototype.hashCode = function () {
        return ((stringHash(this.kind) + this.names.names.hashCode() + this.properties.hashCode()) | 0);
    };
    return ClassType;
}(NamedType));
exports.ClassType = ClassType;
var UnionType = /** @class */ (function (_super) {
    __extends(UnionType, _super);
    function UnionType(names, members) {
        var _this = _super.call(this, "union", names) || this;
        _this.members = members;
        _this.findMember = function (kind) {
            return _this.members.find(function (t) { return t.kind === kind; });
        };
        return _this;
    }
    Object.defineProperty(UnionType.prototype, "children", {
        get: function () {
            return this.members;
        },
        enumerable: true,
        configurable: true
    });
    UnionType.prototype.equals = function (other) {
        if (!(other instanceof UnionType))
            return false;
        return this.names.names.equals(other.names.names) && this.members.equals(other.members);
    };
    UnionType.prototype.hashCode = function () {
        return (stringHash(this.kind) + this.names.names.hashCode() + this.members.hashCode()) | 0;
    };
    return UnionType;
}(NamedType));
exports.UnionType = UnionType;
function removeNullFromUnion(t) {
    if (!t.members.some(isNull)) {
        return [false, t.members];
    }
    return [true, t.members.filterNot(isNull).toOrderedSet()];
}
exports.removeNullFromUnion = removeNullFromUnion;
function nullableFromUnion(t) {
    var _a = removeNullFromUnion(t), hasNull = _a[0], nonNulls = _a[1];
    if (!hasNull)
        return null;
    if (nonNulls.size !== 1)
        return null;
    return Support_1.defined(nonNulls.first());
}
exports.nullableFromUnion = nullableFromUnion;
// FIXME: The outer OrderedSet should be some Collection, but I can't figure out
// which one.  Collection.Indexed doesn't work with OrderedSet, which is unfortunate.
function orderedSetUnion(sets) {
    var setArray = sets.toArray();
    if (setArray.length === 0)
        return immutable_1.OrderedSet();
    if (setArray.length === 1)
        return setArray[0];
    return (_a = setArray[0]).union.apply(_a, setArray.slice(1));
    var _a;
}
function allNamedTypes(graph, childrenOfType) {
    var seen = immutable_1.Set();
    var types = immutable_1.List();
    function addFromType(t) {
        if (seen.has(t))
            return;
        seen = seen.add(t);
        var children = childrenOfType ? childrenOfType(t) : t.children;
        children.forEach(addFromType);
        if (t instanceof ClassType || t instanceof UnionType) {
            types = types.push(t);
        }
    }
    graph.forEach(addFromType);
    return types.reverse().toOrderedSet();
}
exports.allNamedTypes = allNamedTypes;
function splitClassesAndUnions(types) {
    var classes = types
        .filter(function (t) { return t instanceof ClassType; })
        .toOrderedSet();
    var unions = types
        .filter(function (t) { return t instanceof UnionType; })
        .toOrderedSet();
    return { classes: classes, unions: unions };
}
exports.splitClassesAndUnions = splitClassesAndUnions;
function allClassesAndUnions(graph, childrenOfType) {
    var types = allNamedTypes(graph, childrenOfType);
    return splitClassesAndUnions(types);
}
exports.allClassesAndUnions = allClassesAndUnions;
function matchType(t, anyType, nullType, boolType, integerType, doubleType, stringType, arrayType, classType, mapType, unionType) {
    if (t instanceof PrimitiveType) {
        var f = {
            any: anyType,
            null: nullType,
            bool: boolType,
            integer: integerType,
            double: doubleType,
            string: stringType
        }[t.kind];
        if (f)
            return f(t);
        throw "Unknown PrimitiveType: " + t.kind;
    }
    else if (t instanceof ArrayType)
        return arrayType(t);
    else if (t instanceof ClassType)
        return classType(t);
    else if (t instanceof MapType)
        return mapType(t);
    else if (t instanceof UnionType)
        return unionType(t);
    throw "Unknown Type";
}
exports.matchType = matchType;
